@@ -1,10 +1,10 @@
 from .idl import Idl
 from .main import validate, pretty_parse
-from .visitor import Visitor
+from .visitor import IdlNodeVisitor
 
-__all__ = (
+__all__ = [
     'validate',
     'Idl',
     'pretty_parse',
-    'Visitor',
-)
+    'IdlNodeVisitor',
+]
