@@ -1,3 +1,5 @@
+from typing import final
+
 from more_itertools.more import first
 
 from .expr import (
@@ -7,22 +9,25 @@ from .expr import (
     CallbackInterface,
     Const,
     Constructor,
+    Definitions,
     Dictionary,
     Enum,
     ExtendedAttribute,
     Field,
-    Identifier,
-    IdentifierList,
-    Includes,
     IdlType,
+    Includes,
     Infinity,
     Interface,
     InterfaceMixin,
     Iterable_,
     Literal,
+    LiteralList,
     MapLike,
     Namespace,
+    NaN,
+    Null,
     Operation,
+    SetLike,
     Typedef,
     Value,
 )
@@ -30,6 +35,7 @@ from .generated import WebIDLParser, WebIDLParserVisitor
 from .utils import setup_type
 
 
+@final
 class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
 
     def __init__(self, tree):
@@ -42,9 +48,12 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
         return ctx.definitions().accept(self)
 
     def visitDefinitions(self, ctx: WebIDLParser.DefinitionsContext):
-        return [
-            definition.accept(self) for definition in ctx.extendedDefinition()
-        ]
+        return Definitions(
+            definitions=[
+                definition.accept(self)
+                for definition in ctx.extendedDefinition()
+            ],
+        )
 
     def visitExtendedDefinition(
             self, ctx: WebIDLParser.ExtendedDefinitionContext):
@@ -58,7 +67,7 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
     def visitDictionary(self, ctx: WebIDLParser.DictionaryContext):
         return Dictionary(
             partial=ctx.PARTIAL() is not None,
-            name=ctx.IDENTIFIER_WEBIDL().getText(),
+            name=ctx.IDENTIFIER().getText(),
             members=[member.accept(self) for member in ctx.dictionaryMembers()],
             inheritance=ctx.inheritance() and ctx.inheritance().accept(self),
         )
@@ -69,8 +78,7 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
             'typedef-type',
         )
 
-        return Typedef(
-            idl_type=idl_type, name=ctx.IDENTIFIER_WEBIDL().getText())
+        return Typedef(idl_type=idl_type, name=ctx.IDENTIFIER().getText())
 
     def visitIncludesStatement(
             self, ctx: WebIDLParser.IncludesStatementContext):
@@ -79,14 +87,47 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
 
     def visitEnum_(self, ctx: WebIDLParser.Enum_Context):
         return Enum(
-            name=ctx.IDENTIFIER_WEBIDL().getText(),
+            name=ctx.IDENTIFIER().getText(),
             values=[
-                Identifier(
+                Literal(
                     type='enum-value', value=enum_value.getText().strip('"'))
-                for enum_value in ctx.STRING_WEBIDL()
+                for enum_value in ctx.StringLiteral()
             ]
         )
 
+    def visitInterfaceRest(self, ctx: WebIDLParser.InterfaceRestContext):
+        return Interface(
+            name=ctx.IDENTIFIER().getText(),
+            inheritance=ctx.inheritance() and ctx.inheritance().accept(self),
+            members=[member.accept(self) for member in ctx.interfaceMembers()],
+        )
+
+    def visitPartialInterfaceRest(
+            self, ctx: WebIDLParser.PartialInterfaceRestContext):
+        return Interface(
+            name=ctx.IDENTIFIER().getText(),
+            members=[
+                member.accept(self) for member in ctx.partialInterfaceMembers()
+            ],
+        )
+
+    def visitMixinRest(self, ctx: WebIDLParser.MixinRestContext):
+        return InterfaceMixin(
+            name=ctx.IDENTIFIER().getText(),
+            members=[
+                member.accept(self) for member in ctx.mixinMembers()
+            ],
+        )
+
+    def visitNamespace(self, ctx: WebIDLParser.NamespaceContext):
+        return Namespace(
+            name=ctx.IDENTIFIER().getText(),
+            members=[
+                member.accept(self) for member in ctx.namespaceMembers()
+            ],
+            partial=ctx.PARTIAL() is not None,
+        )
+
     def visitPartial(self, ctx: WebIDLParser.PartialContext):
         interface_or_mixin = ctx.partialInterfaceOrPartialMixin().accept(self)
 
@@ -94,89 +135,40 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
 
         return interface_or_mixin
 
-    def visitInterfaceRest(self, ctx: WebIDLParser.InterfaceRestContext):
-        return Interface(
-            name=ctx.IDENTIFIER_WEBIDL().getText(),
-            inheritance=ctx.inheritance() and ctx.inheritance().accept(self),
-            members=[member.accept(self) for member in ctx.interfaceMembers()],
-        )
+    def _member_with_extended_attribute(self, member, extended_attribute):
+        member = member.accept(self)
 
-    def visitPartialInterfaceRest(
-            self, ctx: WebIDLParser.PartialInterfaceRestContext):
-        return Interface(
-            name=ctx.IDENTIFIER_WEBIDL().getText(),
-            members=[
-                member.accept(self) for member in ctx.partialInterfaceMembers()
-            ],
-        )
+        if extended_attribute is not None:
+            member.ext_attrs = extended_attribute.accept(self)
 
-    def visitMixinRest(self, ctx: WebIDLParser.MixinRestContext):
-        return InterfaceMixin(
-            name=ctx.IDENTIFIER_WEBIDL().getText(),
-            members=[
-                member.accept(self) for member in ctx.mixinMembers()
-            ],
-        )
-
-    def visitNamespace(self, ctx: WebIDLParser.NamespaceContext):
-        return Namespace(
-            name=ctx.IDENTIFIER_WEBIDL().getText(),
-            members=[
-                member.accept(self) for member in ctx.namespaceMembers()
-            ],
-            partial=ctx.PARTIAL() is not None,
-        )
+        return member
 
     def visitNamespaceMembers(self, ctx: WebIDLParser.NamespaceMembersContext):
-        member = ctx.namespaceMember().accept(self)
-
-        if extended_attribute := ctx.extendedAttributeList():
-            member.ext_attrs = extended_attribute.accept(self)
-
-        return member
+        return self._member_with_extended_attribute(
+            ctx.namespaceMember(), ctx.extendedAttributeList())
 
     def visitMixinMembers(self, ctx: WebIDLParser.MixinMembersContext):
-        member = ctx.mixinMember().accept(self)
-
-        if extended_attribute := ctx.extendedAttributeList():
-            member.ext_attrs = extended_attribute.accept(self)
-
-        return member
+        return self._member_with_extended_attribute(
+            ctx.mixinMember(), ctx.extendedAttributeList())
 
     def visitPartialInterfaceMembers(
             self, ctx: WebIDLParser.PartialInterfaceMembersContext):
-        member = ctx.partialInterfaceMember().accept(self)
-
-        if extended_attribute := ctx.extendedAttributeList():
-            member.ext_attrs = extended_attribute.accept(self)
-
-        return member
+        return self._member_with_extended_attribute(
+            ctx.partialInterfaceMember(), ctx.extendedAttributeList())
 
     def visitInterfaceMembers(self, ctx: WebIDLParser.InterfaceMembersContext):
-        member = ctx.interfaceMember().accept(self)
-
-        if extended_attribute := ctx.extendedAttributeList():
-            member.ext_attrs = extended_attribute.accept(self)
-
-        return member
+        return self._member_with_extended_attribute(
+            ctx.interfaceMember(), ctx.extendedAttributeList())
 
     def visitDictionaryMembers(
             self, ctx: WebIDLParser.DictionaryMembersContext):
-        member = ctx.dictionaryMember().accept(self)
-
-        if extended_attribute := ctx.extendedAttributeList():
-            member.ext_attrs = extended_attribute.accept(self)
-
-        return member
+        return self._member_with_extended_attribute(
+            ctx.dictionaryMember(), ctx.extendedAttributeList())
 
     def visitCallbackInterfaceMembers(
             self, ctx: WebIDLParser.CallbackInterfaceMembersContext):
-        member = ctx.callbackInterfaceMember().accept(self)
-
-        if extended_attribute := ctx.extendedAttributeList():
-            member.ext_attrs = extended_attribute.accept(self)
-
-        return member
+        return self._member_with_extended_attribute(
+            ctx.callbackInterfaceMember(), ctx.extendedAttributeList())
 
     def visitMixinMember(self, ctx: WebIDLParser.MixinMemberContext):
         if ctx.READONLY() is None:
@@ -200,7 +192,7 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
 
         return Field(
             required=ctx.REQUIRED() is not None,
-            name=ctx.IDENTIFIER_WEBIDL().getText(),
+            name=ctx.IDENTIFIER().getText(),
             default=default,
             idl_type=idl_type,
         )
@@ -214,6 +206,39 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
 
         return attribute
 
+    def visitStringifier(self, ctx: WebIDLParser.StringifierContext):
+        if stringifier := ctx.stringifierRest():
+            stringifier = stringifier.accept(self)
+            stringifier.special = 'stringifier'
+
+            return stringifier
+
+        return Operation(name='', idl_type=None, special='stringifier')
+
+    def visitStringifierRest(self, ctx: WebIDLParser.StringifierRestContext):
+        if regular_operation := ctx.regularOperation():
+            return self._operation(regular_operation)
+
+        attribute = ctx.attributeRest().accept(self)
+        attribute.readonly = ctx.READONLY() is not None
+
+        return attribute
+
+    def visitStaticMember(self, ctx: WebIDLParser.StaticMemberContext):
+        member = ctx.staticMemberRest().accept(self)
+        member.special = 'static'
+
+        return member
+
+    def visitStaticMemberRest(self, ctx: WebIDLParser.StaticMemberRestContext):
+        if regular_operation := ctx.regularOperation():
+            return self._operation(regular_operation)
+
+        attribute = ctx.attributeRest().accept(self)
+        attribute.readonly = ctx.READONLY() is not None
+
+        return attribute
+
     def visitCallbackOrInterfaceOrMixin(
             self, ctx: WebIDLParser.CallbackOrInterfaceOrMixinContext):
         if callback := ctx.callbackRestOrInterface():
@@ -227,7 +252,7 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
             return callback_rest.accept(self)
 
         return CallbackInterface(
-            name=ctx.IDENTIFIER_WEBIDL().getText(),
+            name=ctx.IDENTIFIER().getText(),
             inheritance=None,
             members=[
                 member.accept(self)
@@ -262,6 +287,12 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
             readonly=ctx.READONLY() is not None,
         )
 
+    def visitSetlikeRest(self, ctx: WebIDLParser.SetlikeRestContext):
+        return SetLike(
+            idl_type=[ctx.typeWithExtendedAttributes().accept(self)],
+            readonly=ctx.READONLY() is not None,
+        )
+
     def visitCallbackInterfaceMember(
             self, ctx: WebIDLParser.CallbackInterfaceMemberContext):
         if regular_operation := ctx.regularOperation():
@@ -276,11 +307,11 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
         if float_literal := ctx.floatLiteral():
             return float_literal.accept(self)
 
-        return Identifier(type='number', value=ctx.INTEGER_WEBIDL().getText())
+        return Literal(type='number', value=ctx.IntegerLiteral().getText())
 
     def visitFloatLiteral(self, ctx: WebIDLParser.FloatLiteralContext):
         if ctx.NAN():
-            return Literal(type='NaN')
+            return NaN()
 
         if ctx.INFINITY():
             return Infinity(negative=False)
@@ -288,16 +319,16 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
         if ctx.MINUS_INFINITY():
             return Infinity(negative=True)
 
-        return Identifier(type='number', value=ctx.DECIMAL_WEBIDL().getText())
+        return Literal(type='number', value=ctx.DecimalLiteral().getText())
 
     def visitBooleanLiteral(self, ctx: WebIDLParser.BooleanLiteralContext):
-        return Identifier(type='boolean', value=ctx.getText() == 'true')
+        return Literal(type='boolean', value=ctx.getText() == 'true')
 
     def visitConst_(self, ctx: WebIDLParser.Const_Context):
         return Const(
             idl_type=ctx.constType().accept(self),
             value=ctx.constValue().accept(self),
-            name=ctx.IDENTIFIER_WEBIDL().getText(),
+            name=ctx.IDENTIFIER().getText(),
         )
 
     def visitCallbackRest(self, ctx: WebIDLParser.CallbackRestContext):
@@ -307,10 +338,9 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
         setup_type(idl_type := ctx.returnType().accept(self), 'return-type')
 
         return Callback(
-            name=ctx.IDENTIFIER_WEBIDL().getText(),
+            name=ctx.IDENTIFIER().getText(),
             idl_type=idl_type,
             arguments=arguments or [],
-            ext_attrs=[]
         )
 
     def visitConstructor(self, ctx: WebIDLParser.ConstructorContext):
@@ -388,37 +418,41 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
 
     def visitExtendedAttributeNoArgs(
             self, ctx: WebIDLParser.ExtendedAttributeNoArgsContext):
-        return self._extended_attribute(ctx.IDENTIFIER_WEBIDL().getText())
+        return self._extended_attribute(ctx.IDENTIFIER().getText())
 
     def visitExtendedAttributeNamedArgList(
             self, ctx: WebIDLParser.ExtendedAttributeNamedArgListContext):
         return self._extended_attribute(
-            name=ctx.name.text,
+            name=ctx.name.text,  # type: ignore
             arguments=ctx.argumentList().accept(self),
-            rhs=Identifier(type='identifier', value=ctx.rhs.text),
+            rhs=Literal(type='identifier', value=ctx.rhs.text),  # type: ignore
         )
 
     def visitExtendedAttributeArgList(
             self, ctx: WebIDLParser.ExtendedAttributeArgListContext):
         return self._extended_attribute(
-            name=ctx.name.text, arguments=ctx.argumentList().accept(self))
+            name=ctx.name.text,  # type: ignore
+            arguments=ctx.argumentList().accept(self),
+        )
 
     def visitExtendedAttributeIdentList(
             self, ctx: WebIDLParser.ExtendedAttributeIdentListContext):
         return self._extended_attribute(
-            name=ctx.name.text, rhs=ctx.identifierList().accept(self))
+            name=ctx.name.text,  # type: ignore
+            rhs=ctx.identifierList().accept(self)
+        )
 
     def visitExtendedAttributeIdent(
             self, ctx: WebIDLParser.ExtendedAttributeIdentContext):
         return self._extended_attribute(
-            name=ctx.name.text, rhs=ctx.rhs.accept(self))
+            name=ctx.name.text, rhs=ctx.rhs.accept(self))  # type: ignore
 
     def visitIdentifierList(self, ctx: WebIDLParser.IdentifierListContext):
         identifiers = [
             identifier.accept(self) for identifier in ctx.identifier()
         ]
 
-        return IdentifierList(
+        return LiteralList(
             value=[Value(value=identifier.value) for identifier in identifiers],
             type=f'{first(identifiers).type}-list',
         )
@@ -439,6 +473,7 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
             name=ctx.argumentName().accept(self),
             optional=optional is not None,
             default=default,
+            variadic=ctx.ELLIPSIS() is not None,
         )
 
     def visitReadOnlyMember(self, ctx: WebIDLParser.ReadOnlyMemberContext):
@@ -466,19 +501,19 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
         return ctx.defaultValue().accept(self)
 
     def visitDefaultValue(self, ctx: WebIDLParser.DefaultValueContext):
-        if _str := ctx.STRING_WEBIDL():
-            return Identifier(type='string', value=_str.getText().strip('"'))
+        if str_ := ctx.StringLiteral():
+            return Literal(type='string', value=str_.getText().strip('"'))
 
         if const_value := ctx.constValue():
             return const_value.accept(self)
 
         if ctx.LEFT_BRACE():
-            return Literal(type='dictionary')
+            return Literal(type='dictionary', value=[])
 
         if ctx.NULL():
-            return Literal(type='null')
+            return Null()
 
-        return Identifier(type='sequence', value=[])
+        return Literal(type='sequence', value=[])
 
     def visitType_(self, ctx: WebIDLParser.Type_Context):
         if single_type := ctx.singleType():
@@ -516,7 +551,10 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
             )
 
         return IdlType(
-            idl_type=ctx.unionType().accept(self), nullable=nullable)
+            idl_type=ctx.unionType().accept(self),
+            nullable=nullable,
+            union=True,
+        )
 
     def visitReturnType(self, ctx: WebIDLParser.ReturnTypeContext):
         if void := ctx.VOID():
@@ -549,8 +587,13 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
                 nullable=nullable,
             )
 
+        distinguishable_type = ctx.distinguishableType().accept(self)
+
+        if isinstance(distinguishable_type, IdlType):
+            return distinguishable_type
+
         return IdlType(
-            idl_type=ctx.distinguishableType().accept(self), nullable=nullable)
+            idl_type=distinguishable_type, nullable=nullable)
 
     def visitGenericType(self, ctx: WebIDLParser.GenericTypeContext):
         return ctx.typeWithExtendedAttributes().accept(self)
@@ -566,16 +609,26 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
         if type_ := ctx.bufferRelatedType():
             return type_.accept(self)
 
+        if type_ := ctx.recordType():
+            return type_.accept(self)
+
         return ctx.getChild(0).getText()
 
+    def visitRecordType(self, ctx: WebIDLParser.RecordTypeContext):
+        return IdlType(
+            generic='record',
+            idl_type=[
+                IdlType(idl_type=ctx.stringType().accept(self)),
+                ctx.typeWithExtendedAttributes().accept(self),
+            ]
+        )
+
     def visitConstType(self, ctx: WebIDLParser.ConstTypeContext):
         if idl_type := ctx.primitiveType():
             idl_type = idl_type.accept(self)
 
         return IdlType(
-            type='const-type',
-            idl_type=idl_type or ctx.IDENTIFIER_WEBIDL().getText(),
-        )
+            type='const-type', idl_type=idl_type or ctx.IDENTIFIER().getText())
 
     def visitPrimitiveType(self, ctx: WebIDLParser.PrimitiveTypeContext):
         if type_ := ctx.unsignedIntegerType():
@@ -599,6 +652,9 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
         return ctx.returnType().accept(self)
 
     def visitIntegerType(self, ctx: WebIDLParser.IntegerTypeContext):
+        if long := ' '.join(long.getText() for long in ctx.LONG()):
+            return long
+
         return ctx.getText()
 
     def visitFloatType(self, ctx: WebIDLParser.FloatTypeContext):
@@ -617,16 +673,16 @@ class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods
         return ctx.getText()
 
     def visitInheritance(self, ctx: WebIDLParser.InheritanceContext):
-        return ctx.IDENTIFIER_WEBIDL().getText()
+        return ctx.IDENTIFIER().getText()
 
     def visitOther(self, ctx: WebIDLParser.OtherContext):
-        if ctx.INTEGER_WEBIDL():
+        if ctx.IntegerLiteral():
             type_ = 'integer'
-        elif ctx.DECIMAL_WEBIDL():
+        elif ctx.DecimalLiteral():
             type_ = 'decimal'
-        elif ctx.STRING_WEBIDL():
+        elif ctx.StringLiteral():
             type_ = 'string'
         else:
             type_ = 'identifier'
 
-        return Identifier(type=type_, value=ctx.getText())
+        return Literal(type=type_, value=ctx.getText())
