@@ -8,16 +8,15 @@ from .idl import Idl
 
 
 def validate(file: str):
-    if errors := Idl(file).validate():
-        return errors
-
-    return 'Ok'
+    return Idl(file).validate()
 
 
 def pretty_parse(file: str):
+    ast = Visitor(Idl(file).parse()).run()
+
     return [
         attr.asdict(definition)
-        for definition in Visitor(Idl(file).parse()).run()
+        for definition in ast.definitions
     ]
 
 
